#include "pvo-utils.hh"
#include <time.h>
#include <stdlib.h>
#include <math.h>
#include <string>
#include <glob.h>
#include <iostream>
#include <sstream>
#include <iomanip>

namespace PVO
{

  std::string getPSRFITSdefn(){ 
    return getenv("PSRFITSDEFN"); 
  }

  void PVOError(std::string subroutine, std::string error)
  {
    /**
     * Prints an ERROR message to the standard error stream, quoting
     * the subroutine that the problem occurred in, a descriptive
     * warning text, and sounding the bell.
     *
     * Format: ERROR \<function\> : Something bad happened.
     */
    std::string preamble = "\aERROR <" + subroutine + "> : ";
    std::cerr << preamble;
    for(unsigned int i=0;i<error.size();i++){
      std::cerr << error[i];
      if((i!=error.size()-1)&&(error[i]=='\n'))
	std::cerr << std::setw(preamble.size()-1) <<": ";
    }
  }
  //------------------------------------------------------------

  std::vector<std::string> getAllFitsFiles()
  {
    std::vector<std::string> namelist;
    glob_t list1,list2,list3;
    unsigned int i;

    list1.gl_offs = 0;
    list2.gl_offs = 0;
    list3.gl_offs = 0;

    glob("*.fits", GLOB_DOOFFS, NULL, &list1);
    glob("*.rf", GLOB_DOOFFS, NULL, &list2);
    glob("*.sf", GLOB_DOOFFS, NULL, &list3);

    namelist.resize(list1.gl_pathc+list2.gl_pathc+list3.gl_pathc);
    for(i=0;i<list1.gl_pathc+list1.gl_offs;i++) 
      namelist[i] = list1.gl_pathv[i];
    for(i=0;i<list2.gl_pathc+list2.gl_offs;i++) 
      namelist[list1.gl_pathc+i] = list2.gl_pathv[i];
    for(i=0;i<list3.gl_pathc+list3.gl_offs;i++) 
      namelist[list1.gl_pathc+list2.gl_pathc+i] = list3.gl_pathv[i];
    
    return namelist;
  }

  //------------------------------------------------------------

  std::string removeLeadingBlanks(std::string s)
  {
    /**
     * All blank spaces from the start of the string to the first
     * non-blank-space character are deleted.
     */ 
    int i=0;
    while(s[i]==' '){
      i++;
    }
    std::string newstring;
    for(unsigned int j=i;j<s.size();j++) newstring += s[j];
    return newstring;
  }
  //------------------------------------------------------------------------

  std::string deblank(std::string s)
  {
    /**
     * All blank spaces from the start of the string to the first
     * non-blank-space character, and from the last non-blank-space
     * character to the end are deleted.
     */ 
    int beg=0;
    while(s[beg]==' '){
      beg++;
    }
    int end=s.size()-1;
    while(s[end]==' '){
      end--;
    }
    std::string newstring;
    for(int j=beg;j<=end;j++) newstring += s[j];
    return newstring;
  }
  //------------------------------------------------------------------------

  std::string charToBinary(char c)
  {
    /**
     * A single byte, represented by a char type, is read in and converted
     * to a binary representation.
     *
     * Used for debugging the single-bit writing and reading tasks.
     *
     * \param c A single byte, of char type.
     * \return A std:string that shows the binary representation of that byte.
     */

    std::stringstream ss;
    int num=c;
    if (num<0) num+=256;
    int factor = 128;
    for(int i=0;i<8;i++){
      ss << num/factor;
      num -= num/factor * factor;
      factor /= 2;
    }
    return ss.str();
  }
  //------------------------------------------------------------------------

  std::string binaryToString(std::vector<bool> v)
  {
    /**
     * A string of bits, read in as a vector of bool types, is coverted to the
     *  equivalent string of bytes, in the form of chars (in fact, a std::string).
     *
     * Used for debugging the single-bit writing and reading tasks.
     *
     * \param v String of bits.
     * \return Std::string of bytes (each byte a char).
     */
    int size = v.size();
    int nbytes = (size-1)/8 + 1;
    std::string output;
    for(int byte=0;byte<nbytes;byte++){
      int c=0;
      for(int bit=0;bit<8;bit++){
	int pt = byte*8+bit;
	if(v[pt]) c += int(pow(2,7-bit));
      }
      //    std::cerr << int(c) << std::endl;
      output += char(c);
    }
    return output;
  }
  //------------------------------------------------------------------------

  unsigned int * binaryToInt(std::vector<bool> v)
  {
    /**
     * A string of bits, read in as a vector of bool types, is coverted to a 
     *  series of bytes, which are expressed as an array of integers (with the
     *  decimal value of each byte).
     *
     * Used for debugging the single-bit writing and reading tasks.
     *
     * \param v String of bits.
     * \return Sequence of integers, being the decimal value of each byte.
     */
    int size = v.size();
    int nbytes = (size-1)/8 + 1;
    unsigned int *output = new unsigned int[nbytes];
    for(int byte=0;byte<nbytes;byte++){
      int c=0;
      for(int bit=0;bit<8;bit++){
	int pt = byte*8+bit;
	if(v[pt]) c += (unsigned int)pow(2,7-bit);
      }
      //    std::cerr << c << std::endl;
      output[byte] = c;
    }
    return output;
  }
  //------------------------------------------------------------------------

  double dmsToDec(std::string input)
  {
    /** 
     *   Assumes the angle given is in degrees, so if passing RA as
     *   the argument, need to multiply by 15 to get the result in
     *   degrees rather than hours.  The sign of the angle is
     *   preserved, if present.
     */

    std::string dms = removeLeadingBlanks(input);

    bool isNeg = false;
    if(dms[0]=='-') isNeg = true;

    for(unsigned int i=0;i<dms.size();i++) if(dms[i]==':') dms[i]=' ';

    std::stringstream ss;
    ss.str(dms);

    std::string deg,min,sec;
    getline(ss,deg,' ');
    getline(ss,min,' ');
    getline(ss,sec);
    char *end;
    double d = strtod(deg.c_str(),&end);
    double m = strtod(min.c_str(),&end);
    double s = strtod(sec.c_str(),&end);  

    double dec = fabs(d) + m/60. + s/3600.;
    if(isNeg) dec = dec * -1.;

    return dec;

  }
  //------------------------------------------------------------------------

  std::string decToDMS(double input, int secondPrecision, std::string separator)
  {
    /**
     *   This is the general form, where one can specify the degree of
     *   precision of the seconds, and the separating character.
     *  \param input Angle in decimal degrees.
     *  \param secondPrecision How many decimal places to quote the seconds to.
     *  \param separator The character (or string) to use as a
     *  separator between hh and mm, and mm and ss.sss.
     */ 
    double hour = trunc(input);
    double fraction = fabs(input - hour);
    double min = trunc( fraction * 60. );
    fraction = fraction * 60. - min;
    double sec = fraction * 60.;
  
    int secondWidth = 2;
    if(secondPrecision>0) secondWidth += 1 + secondPrecision;

    std::stringstream output;
    output.setf(std::ios::fixed);
    output << std::setw(2) << std::setfill('0') << std::setprecision(0)
	   << hour << separator 
	   << std::setw(2) << std::setfill('0') << std::setprecision(0)
	   << min  << separator ;
    output << std::setw(secondWidth) << std::setfill('0') 
	   << std::setprecision(secondPrecision) << sec;

    return output.str();
  }
  //------------------------------------------------------------------------

  std::string getIAUName(double ra, double dec)
  {
    /**
     *  The positions are assumed to be J2000.0, since the name is
     *  prefixed by a J.
     *
     *  \param ra RA position, in degrees.
     *  \param dec Dec position, in degrees.
     *  \return Name of the form J1234m4321 or J1234p0123
     */

    double raHrs = ra / 15.;
    int h = int(raHrs);
    int m = (int)(fmod(raHrs,1.)*60.);
    int s = (int)(fmod(raHrs,1./60.)*3600.);
    std::stringstream ss(std::stringstream::out);
    ss.setf(std::ios::showpoint);
    ss.setf(std::ios::fixed);
    ss << "J"; 
    ss<<std::setw(2)<<std::setfill('0')<<h;
    ss<<std::setw(2)<<std::setfill('0')<<m;
    ss<<std::setw(2)<<std::setfill('0')<<s;
    int sign = int( dec / fabs(dec) );
    double d = dec / sign;
    h = int(d);
    m = (int)(fmod(d,1.)*60.);
    s = (int)(fmod(d,1./60.)*3600.);
    if(sign==1) ss<<"p"; else ss<<"m";
    ss<<std::setw(2)<<std::setfill('0')<<h;
    ss.unsetf(std::ios::showpos);
    ss<<std::setw(2)<<std::setfill('0')<<m;
    ss<<std::setw(2)<<std::setfill('0')<<s;
    return ss.str();
  }
  //------------------------------------------------------------------------

  time_t my_timegm (struct tm *tm) {
    /**
     * Local function to get the proper GM time.
     */
    time_t ret;
    char *tz;

    tz = getenv("TZ");
    setenv("TZ", "", 1);
    tzset();
    ret = mktime(tm);
    if (tz) setenv("TZ", tz, 1);
    else    unsetenv("TZ");
    tzset();
    return ret;
  }

  std::string addDay(std::string date)
  {
    /**
     *   This assumes the format is yy/mm/dd or yymmdd or yy-mm-dd
     *    -- the crucial thing is that the day number is in the last spot.
     *   The format of the date is preserved, as long as it is one of those 3.
     *  \param date Input date.
     *  \return Date of same format as input, but one day later.
     */ 

    struct tm time;
    time.tm_hour=0;
    time.tm_min =0;
    time.tm_sec =1;
    char buf[255];
    std::string output;
    time_t temp;
    if(date.size()==6) {
      strptime(date.c_str(), "%y%m%d", &time);
      time.tm_mday += 1;
      temp = my_timegm(&time);
      strftime(buf, sizeof(buf), "%y%m%d", gmtime(&temp));
    }
    else if(date[2]=='/'){
      strptime(date.c_str(), "%y/%m/%d", &time);
      time.tm_mday += 1;
      temp = my_timegm(&time);
      strftime(buf, sizeof(buf), "%y/%m/%d", gmtime(&temp));
    }
    else if(date[2]=='-'){
      strptime(date.c_str(), "%y-%m-%d", &time);
      time.tm_mday += 1;
      temp = my_timegm(&time);
      strftime(buf, sizeof(buf), "%y-%m-%d", gmtime(&temp));
    }
    else{
      std::cerr << "<addDay> : date (" << date << ") has unknown format!\n";
      return date;
    }
    output = buf;
    return output;
  
  }
  //------------------------------------------------------------------------

  std::string makeDateObs(std::string date, std::string time)
  {

    /**
     * This produces a string of the form YY-MM-DDThh:mm:ss.sss, using
     * strings containing the date and the time. It is simply a matter
     * of concatenating them, but we check the date string to make sure
     * it has the hyphens in it.
     *
     * \param date A date string, assumed to be of the format YY-MM-DD,
     * YY/MM/DD, or YYMMDD.
     * \param time A time string, assumed to be in the format
     * hh:mm:ss[.sss] (the fractional seconds are optional).
     * \return A properly formatted date/time string.
     */

    std::string date_obs;
    bool hashyphens = false;
    for(unsigned int i=0;i<date.size();i++) 
      hashyphens = hashyphens || (date[i]=='-') || (date[i] == '/');
    if(hashyphens) date_obs = date;
    else date_obs = date.substr(0,2) + "-" + 
	   date.substr(2,2) + "-" + date.substr(4,2);
    if(date[0]=='0') date_obs = "20" + date_obs;
    else             date_obs = "19" + date_obs;
    for(unsigned int i=0;i<date_obs.size();i++) 
      if(date_obs[i]=='/') date_obs[i]='-';

    date_obs += "T" + time;

    return date_obs;

  }
  //------------------------------------------------------------------------

  double angSep(std::string ra1, std::string dec1, 
		std::string ra2, std::string dec2)
  {
    /**
     *  Calculates the angular separation between two sky positions,
     *  given as strings for RA and DEC. Uses the function
     *  angSep(double,double,double,double).
     * \param ra1 The right ascension for the first position.
     * \param dec1 The declination for the first position.
     * \param ra2 The right ascension for the second position.
     * \param dec2 The declination for the second position.
     * \return The angular separation in degrees.
     */
    if((ra1==ra2)&&(dec1==dec2)) 
      return 0.;
    else
      return angSep(dmsToDec(ra1)*15.,dmsToDec(dec1),
		    dmsToDec(ra2)*15.,dmsToDec(dec2) );
  }

  double angSep(double ra1, double dec1, double ra2, double dec2)
  {
    /**
     *  Calculates the angular separation between two sky positions,
     *  where RA and DEC are given in decimal degrees.
     * \param ra1 The right ascension for the first position.
     * \param dec1 The declination for the first position.
     * \param ra2 The right ascension for the second position.
     * \param dec2 The declination for the second position.
     * \return The angular separation in degrees.
     */
    double r1,d1,r2,d2;
    r1 = ra1  * M_PI / 180.;
    d1 = dec1 * M_PI / 180.;
    r2 = ra2  * M_PI / 180.;
    d2 = dec2 * M_PI / 180.;
    long double angsep = cos(r1-r2)*cos(d1)*cos(d2) + sin(d1)*sin(d2);
    
    return acosl(angsep)*180./M_PI;
    
  }
  //------------------------------------------------------------------------

  void equatToGal(double ra, double dec, double &gl, double &gb)
  {
    /**
     * Converts an equatorial (ra,dec) position to galactic
     * coordinates. The equatorial position is assumed to be J2000.0.
     *
     * \param ra Right Ascension, J2000.0
     * \param dec Declination, J2000.0
     * \param gl Galactic Longitude. Returned value.
     * \param gb Galactic Latitude. Returned value.
     */

    const double NGP_RA = 192.859508 * M_PI/180.;
    const double NGP_DEC= 27.128336 * M_PI/180.;
    const double ASC_NODE=32.932;

    double deltaRA = ra * M_PI/180. - NGP_RA;
    double d = dec     * M_PI/180.;
    double sinb = cos(d) * cos(NGP_DEC) * cos(deltaRA) + sin(d) * sin(NGP_DEC);
    gb = asin(sinb);

    // The l in sinl and cosl here is really gl-ASC_NODE
    double sinl = (sin(d)*cos(NGP_DEC) - cos(d)*cos(deltaRA)*sin(NGP_DEC)) / cos(gb);
    double cosl = cos(d) * sin(deltaRA) / cos(gb);
    // atan returns a value between -90 and 90 degrees.
    // Need to correct the value of l according to the correct quandrant it is in.
    // This is worked out using the signs of sinl and cosl
    if(sinl>0){
      if(cosl>0) gl = atan(sinl/cosl);
      else       gl = atan(sinl/cosl) + M_PI;
    }
    else{
      if(cosl>0) gl = atan(sinl/cosl) + 2.*M_PI;
      else       gl = atan(sinl/cosl) + M_PI;
    }

    // Find the correct values of the lat & lon in degrees.
    gb = asin(sinb) * 180. / M_PI;
    gl = gl * 180. / M_PI + ASC_NODE;
  }

}
