#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <string>
#include <unistd.h>
#include <getopt.h>
#include "pvo.hh"
#include "pvo-block.hh"
#include "pvo-header.hh"
#include "pvo-FITS.hh"
#include "pvo-utils.hh"
#include "pvo-metadata.hh"

using namespace PVO;
using namespace PVO::Header;
using namespace PVO::Block;
using namespace PVO::FITS;
using namespace PVO::Meta;
using std::string;


int main(int argc, char * argv[])
{
  /** 
   * PVOSummary is a function to print out basic information for
   * PSRFITS files.
   *
   * This program is a command-line program that looks at PSRFITS
   * files (i.e. files that have extensions .fits, .rf or .sf) and
   * returns a summary of their positions, beam sizes and observation
   * times.
   *
   * The format of usage is: <br> 
   * <tt>> pvosummary </tt><br>
   * to analyse all PSRFITS files in the current directory, or <br>
   * <tt>> pvosummary FILE</tt><br>
   * to analyse the file FILE (more than one file can be given). 
   * Entering<br>
   * <tt>> pvosummary -h</tt><br>
   * will give a help message, using PVO::pvosummary_usage_text.
   *
   * A fuller list of parameters can be written to a csv
   * (comma-separated-values) format file by using the <tt>-c</tt>
   * option, eg:<br>
   * <tt>> pvosummary -c summary.csv</tt>, or <br>
   * <tt>> pvosummary -c summary.csv FILE</tt><br>
   * The list of parameters printed to the csv file can be seen
   * with:<br>
   * <tt>> pvosummary -p</tt><br>
   *
   * The output is a list of valid files, their central RA & DEC,
   * their beam size, the central observing frequency, the requested
   * scan length, and the DATE-OBS keyword (time/date of observation),
   * 
   */


  bool verboseflag = false;
  bool makeCSV = false;
  bool quietMode = false;
  std::string csvFile;

  std::vector<std::string> namelist;
  unsigned int i;

  int c;
  while (true) {
    int option_index = 0;
    static struct option long_options[] = {
      {"help", 0, 0, 'h'},
      {"csv", 1, 0, 'c'},
      {"verbose", 1, 0, 'V'},
      {"version", 1, 0, 'v'},
      {0, 0, 0, 0}
    };

    c = getopt_long (argc, argv, "c:hpqvV",
		     long_options, &option_index);

    if (c == -1) break;
    
    switch (c) {               
    case 0:
      // This is where we deal with -- options that don't have a
      // single letter equivalent
      break;
    case 'c': // make a csv file.
      makeCSV = true;
      csvFile = optarg;
      break;
    case 'h': // give help statement
      std::cout << pvosummary_usage_text;
      return 0;
      break;
    case 'p': // list the parameters
      std::cout << Meta::metadataList;
      return 0;
      break;
    case 'v': // return version number
      std::cout << PVO::PVOSUMMARY << " version " << PVO::PVO_VERSION << "\n";
      return 0;
      break;
    case 'V': // use verbose mode -- full verbosity for error messages
      verboseflag = true;
      break;
    case 'q': // quiet mode -- nothing written to screen
      quietMode = true;
      break;
    case '?': // invalid option
    case ':': // missing an argument
      std::cout << pvosummary_usage_text_short;
      return 1;
      break;
    default:
      std::cerr << "ERROR: getopt returned character code " <<  c << "\n";
      break;
    }

  }

  if (optind < argc) { 
    // There is a non-option argument -- must be the file to access
    while (optind < argc) namelist.push_back( argv[optind++] );
  }
  else{ 
    //No file has been specified, so we are working on the directory
    //as a whole.
    namelist = PVO::getAllFitsFiles();
    std::cerr << namelist.size() << "\n";
    for(i=0;i<namelist.size();i++) std::cerr << namelist[i] << "\n";
  }

  std::vector<Metadata> goodlist;

  for(i=0;i<namelist.size();i++){

    PSRFits fitsfile;
    fitsfile.setVerbosity(verboseflag);
    int status = fitsfile.openFile(namelist[i]);
  
    if(status==0){
      Metadata newmatch;
      newmatch.define(fitsfile);
      goodlist.push_back(newmatch);
    }

    fitsfile.closeFile();

  }

  if(!quietMode) printMetadataShort(std::cout,goodlist);

  if(makeCSV) printMetadataFull(csvFile, goodlist);

}
